//////////////////////////////////////////////////////////////////////
//
//  _DCPICK.PRG
//
//  Copyright:
//     Alaska Software Inc., (c) 1997. All rights reserved.
//
//  Contents:
//     Class XbpPicklist which displays two listboxes for selection of
//     data. Available data is displayed in the left listbox, while
//     selected data is displayed in the right listbox. Four pushbuttons
//     are used to transfer data between both listboxes.
//
//     The function Picklist() creates a modal dialog for easy use of the
//     XbpPicklist class.
//
//
//  Remarks:
//     The dialog is sizable and calculates position and size of its
//     contained controls when the user changes the size of the window.
//
//////////////////////////////////////////////////////////////////////


#include "Appevent.ch"
#include "Common.ch"
#include "Gra.ch"
#include "Xbp.ch"
#INCLUDE "dcdialog.CH"

/*
 * Picklist class with two listboxes for selection of data.
 * Four pushbuttons are used to transfer items between left and right
 * listbox. The pushbuttons are automatically positioned by a
 * GroupManager object
 */
CLASS DC_XbpPickList FROM XbpStatic, DataRef

   EXPORTED:
   VAR groupMgr
   VAR leftText
   VAR rightText
   VAR leftListBox
   VAR rightListBox
   VAR aSelectedItems
   VAR bLink
   VAR copyitems
   VAR immediate
   VAR oButtonLRone
   VAR oButtonLRall
   VAR oButtonRLone
   VAR oButtonRLall
   VAR oButtonOK
   VAR lEnabled
   VAR whenBlock
   VAR hideBlock
   VAR color
   VAR captionBlock
   VAR getList
   VAR getListPointer
   VAR userCargo
   VAR picture
   VAR optionArray
   VAR leftAnchor

   METHOD init, create
   METHOD setSize, minSize
   METHOD checkButtonState
   METHOD transferItems
   METHOD setData, getData
   METHOD Keyboard
   METHOD enable, disable
   METHOD WhenHideEval
   METHOD refresh

ENDCLASS


/*
 * Initialize listboxes und and group manager
 */
METHOD DC_XbpPickList:init( oParent, oOwner, aPos, aSize, aPP, lVisible, oGetList )

LOCAL aGetList, aGetListItem, aGetOptions, xOptions, bBlock

::getList := oGetList
::getListPointer := ::getList:getListPointer
aGetList := ::getList:getListArray
aGetListItem := ::getList:getListArray[::getListPointer]
aGetOptions := ::getList:getOptions

xOptions := aGetListItem[xGETLIST_OPTIONS]

::whenBlock := DC_GetBlock(aGetListItem[bGETLIST_WHEN])
::hideBlock := DC_GetBlock(aGetListItem[bGETLIST_HIDE])

   ::xbpStatic:init( oParent, oOwner, aPos, aSize, aPP, lVisible )
   ::xbpStatic:type := XBPSTATIC_TYPE_RAISEDBOX
   ::xbpStatic:clipSiblings :=.T.

   ::groupMgr            :=  GroupManager():new( self, self, {35,24} )
   ::groupMgr:horizontal := .F.

   ::leftListBox         := XbpListBox():new( self )

   ::leftText            := XbpStatic():new( self, , {0,0}, {0 ,16} )
   ::leftText:caption    := "Available items"
   ::leftText:type       := XBPSTATIC_TYPE_TEXT
   ::leftText:options    := XBPSTATIC_TEXT_VCENTER+XBPSTATIC_TEXT_LEFT
   ::leftText:tabStop    := .t.
   ::leftText:cargo      := ::leftText:caption

   ::rightListBox        := XbpListBox():new( self )

   ::rightText           := XbpStatic():new( self, , {0,0}, {0 ,16} )
   ::rightText:caption   := "Selected items"
   ::rightText:type      := XBPSTATIC_TYPE_TEXT
   ::rightText:options   := XBPSTATIC_TEXT_VCENTER+XBPSTATIC_TEXT_LEFT
   ::rightText:tabStop   := .t.
   ::rightText:cargo     := ::rightText:caption

   ::leftListBox:keyboard := {|a,b,o|::KeyBoard(a,1)}
   ::leftListBox:lbDblClick := {|a,b,o|::transferItems( ::leftListBox, ::rightListBox, .F.,a,b,o), ;
                                       DC_GetWhen(aGetList) }
   ::leftListBox:multiSelection := .T.
   ::leftListBox:tabStop := .t.

   ::rightListBox:keyboard := {|a,b,o|::KeyBoard(a,2)}
   ::rightListBox:lbDblClick := {|a,b,o|::transferItems( ::rightListBox, ::leftListBox, .F.,a,b,o), ;
                                        DC_GetWhen(aGetList) }
   ::rightListBox:multiSelection := .T.
   ::rightListBox:tabStop := .t.

   ::immediate := .f.

   ::bLink := {||.t.}
   ::copyItems := .f.
   ::lEnabled := .t.

    IF !Empty(aGetListItem[cGETLIST_CAPTION])
      ::title := aGetListItem[cGETLIST_CAPTION]
    ENDIF

    DEFAULT xOptions := {}
    ASize(xOptions,4)

    IF Valtype( xOptions[1] ) == "C"
      ::leftText:caption := xOptions[1]
      ::leftText:cargo := ::leftText:caption
    ENDIF

    IF Valtype( xOptions[2] ) == "C"
      ::rightText:caption := xOptions[2]
      ::rightText:cargo := ::rightText:caption
    ENDIF

    IF Valtype( xOptions[3] ) == 'L'
      ::copyitems := xOptions[3]
    ENDIF

    IF Valtype( xOptions[4] ) == 'L'
      ::immediate := xOptions[4]
    ENDIF

    ::leftAnchor := DC_GetBlock(aGetListItem[bGETLIST_REFVAR]) // left listbox
    ::datalink := DC_GetBlock(aGetListItem[bGETLIST_VAR]) // right listbox

    ::bLink := DC_GetBlock(aGetListItem[bGETLIST_ACTION]) // OK button action

    ::cargo := ;
         { ::getListPointer,                           ; // 1 Pointer
           aGetList,                                   ; // 2 GetList
           aGetListItem[xGETLIST_CARGO] }                // 3 User Cargo

    bBlock := DC_GetBlock(aGetListItem[bGETLIST_GROUP],'O')
    IF Valtype( bBlock ) = 'B'
      Eval( bBlock, self )
    ENDIF

    bBlock := DC_GetBlock(aGetListItem[bGETLIST_PREEVAL])
    IF Valtype(bBlock) = 'B'
      Eval( bBlock, self )
    ENDIF

RETURN self

/*
 * Add four buttons and create the object
 */
METHOD DC_XbpPickList:create( oParent, oOwner, aPos, aSize, aPP, lVisible, lOk )

LOCAL aGetList, aGetListItem, aGetOptions, xOptions, aArray

aGetList := ::getList:getListArray
aGetListItem := ::getList:getListArray[::getListPointer]
aGetOptions := ::getList:getOptions

   lOk := IIF(Valtype(lOk)='L',lOk, !::immediate)

   ::xbpStatic:create( oParent, oOwner, aPos, aSize, aPP, lVisible )

   ::groupMgr:create()

   ::leftListBox:create()
   ::leftText:create()

   ::rightListBox:create()
   ::rightText:create()

   ::oButtonLRone := ::groupMgr:createButton( ">" , {|| ::transferItems( ::leftListBox, ::rightListBox, .F. )  }, .t., .t. )
   ::oButtonLRall := ::groupMgr:createButton( ">>", {|| ::transferItems( ::leftListBox, ::rightListBox, .T. )  }, .t., .t. )
   ::oButtonRLone := ::groupMgr:createButton( "<" , {|| ::transferItems( ::rightListBox, ::leftListBox, .F. )  }, .t., .t. )
   ::oButtonRLall := ::groupMgr:createButton( "<<", {|| ::transferItems( ::rightListBox, ::leftListBox, .T. )  }, .t., .t. )

   IF lOk
     DEFAULT ::bLink := {||.t.}
     ::oButtonOK := ::groupMgr:createButton( "~OK", {||::GetData(),Eval(::bLink)}, .t., .F. )
   ENDIF

   ::setSize( ::xbpStatic:currentSize() )

    IF Valtype(::hideBlock)=='B' .AND. Eval(::hideBlock )
      ::hide()
    ENDIF

    aArray := Eval( ::leftAnchor )

    ::setData( aArray, Eval(::datalink) )
    IF !Empty(aGetListItem[cGETLIST_FONT])
      ::leftListBox:SetFontCompoundName(Alltrim(aGetListItem[cGETLIST_FONT]))
      ::rightListBox:SetFontCompoundName(Alltrim(aGetListItem[cGETLIST_FONT]))
    ENDIF

   ::refresh()

RETURN self

/*
 * Change size of XbpStatic and reposition contained controls
 */
METHOD DC_XbpPickList:setSize( aSize )

   LOCAL aLbSize[2], aPos

   ::xbpStatic:setSize( aSize )

  /*
   * Calculate width and height of the listboxes
   */
   aLbSize[1] := aSize[1] - 2 - 12 - 28 - 12 - 2
   aLbSize[1] /= 2
   aLbSize[2] := aSize[2] - 2 - 16 - 2 - 2

   aPos := {2,2}
   ::leftListBox:setPos ( aPos )
   ::leftListBox:setSize( aLbSize )

   ::leftText:setPos ( { aPos[1], 2+aLbSize[2]+2} )
   ::leftText:setSize( {aLbSize[1],16} )


  /*
   * Calculate position of the four pushbuttons and let the
   * GroupManager center them
   */
   aPos[1] += ( aLbSize[1] + 12 )

   ::groupMgr:setPos( aPos )
   ::groupMgr:center( aSize[2]-2-16, 2 )

  /*
   * Position of the right listbox
   */
   aPos[1] += ( 28 + 12 )

   ::rightListBox:setPos( aPos )
   ::rightListBox:setSize( aLbSize )

   ::rightText:setPos ( { aPos[1], 2+aLbSize[2]+2} )
   ::rightText:setSize( {aLbSize[1],16} )

RETURN self


/*
 * Minimum size of an XbpPicklist object
 */
METHOD DC_XbpPickList:minSize( )
   LOCAL aSize := ::groupMgr:minSize()

   aSize[1] += (  2 * (40 + 12 + 2) )
   aSize[2] += ( 16 + 2 + 2 )

RETURN aSize



/*
 * Toggle disabled/enabled state of pushbuttons
 */
METHOD DC_XbpPickList:checkButtonState
   LOCAL nCount1 := ::leftListBox:numItems()
   LOCAL nCount2 := ::rightListbox:numItems()

   IF nCount1 > 0
      ::groupMgr:part(1):enable()
   ELSE
      ::groupMgr:part(1):disable()
   ENDIF

   IF nCount1 > 1
      ::groupMgr:part(2):enable()
   ELSE
      ::groupMgr:part(2):disable()
   ENDIF

   IF nCount2 > 0
      ::groupMgr:part(3):enable()
   ELSE
      ::groupMgr:part(3):disable()
   ENDIF

   IF nCount2 > 1
      ::groupMgr:part(4):enable()
   ELSE
      ::groupMgr:part(4):disable()
   ENDIF

RETURN self



/*
 * Transfer marked items or all items between both listboxes
 */
METHOD DC_XbpPickList:transferItems( oSender, oReceiver, lAll, a, b, o )

   LOCAL aBuffer, aItem, nItem, i, nCount

   DEFAULT oSender   TO ::leftListBox  , ;
           oReceiver TO ::rightListbox , ;
           lAll      TO .F.

   IF !::lEnabled .OR. ;
      (Valtype(a)=='A' .AND. Valtype(o)=='O' .AND. a[1] >= o:currentSize()[1]-20)
     RETURN self
   ENDIF

   IF ! lAll
      aItem := oSender:getData()
   ELSE
      aItem := AEval( Array(oSender:numItems()), {|x,i| x:=i },,, .T. )
   ENDIF

   nCount := 0

   IF lAll

      nCount  := oSender:numItems()
      aBuffer := AClone( oSender:cargo )
      ASize( oSender:cargo, 0 )
      oSender:clear()

   ELSEIF ! Empty( aItem )

      nCount  := Len( aItem )

     /*
      * Remove items from bottom to top. Otherwise their ordinal
      * position changes.
      */
      aBuffer := Array( nCount )
      FOR i := nCount TO 1 STEP -1
          nItem      := aItem[i]
          aBuffer[i] := oSender:cargo[ nItem ]
          ADel( oSender:cargo, nItem )
          oSender:delItem( nItem )
      NEXT
      ASize( oSender:cargo, oSender:numItems() )
      aItem[1] := Min( aItem[1], oSender:numItems() )
      oSender:setData( {aItem[1]}, .T. )
   ENDIF

  /*
   * Find insertion point
   */
   aItem := oReceiver:getData()
   IF ! Empty( aItem )
      oReceiver:setData( aItem, .F. )
      nItem := ATail( aItem )
   ELSE
      nItem := 0
   ENDIF

  /*
   * insert after last marked element
   */
   FOR i:=1 TO nCount
      nItem ++

      AAdd( oReceiver:cargo, NIL )              // add element to data array

      IF nItem > oReceiver:numItems()
         oReceiver:addItem( aBuffer[i] )        // add item to listbox
      ELSE
         AIns( oReceiver:cargo, nItem )         // insert string to data array
         oReceiver:insItem( nItem, aBuffer[i] ) // insert item to listbox
      ENDIF

      oReceiver:cargo[ nItem ] := aBuffer[i]
   NEXT

   oReceiver:setData( { nItem }, .T. )

   IF ::immediate
      ::GetData()
   ENDIF

   ::checkButtonState()

   IF lAll
     IF !::oButtonLRall:isEnabled()
       SetAppFocus(::oButtonRLAll)
     ELSE
       SetAppFocus(::obuttonLRAll)
     ENDIF
   ENDIF

   ::refresh()

RETURN self



/*
 * Transfer array to left listbox and clear right listbox
 */
METHOD DC_XbpPickList:setdata( aLeftArray, aRightArray )

   aRightArray := IIF( Valtype(aRightArray)='A',aRightArray,{} )

   IF aLeftArray == NIL
      aLeftArray := Eval( ::leftAnchor )
   ENDIF

   ::leftListBox:clear()
   ::leftListBox:cargo := AClone( aLeftArray )
   AEval( aLeftArray, {|c| ::leftListBox:addItem( c ) } )

   ::rightListBox:clear()
   ::rightListBox:cargo := AClone( aRightArray )
   AEval( aRightArray, {|c| ::rightListBox:addItem( c ) } )

   ::checkButtonState()

RETURN



/*
 * Clone items listed in right listbox and eval datalink
 */
METHOD DC_XbpPickList:getData()

LOCAL aRightListBoxItems := AClone( ::rightListBox:cargo )
LOCAL aLeftListBoxItems := AClone( ::leftListBox:cargo )
IIf( ::leftAnchor <> NIL, Eval( ::leftAnchor, aLeftListBoxItems ), aLeftListBoxItems )

RETURN IIf( ::dataLink <> NIL, Eval( ::dataLink, aRightListBoxItems ), aRightListBoxItems )

* -----------

METHOD DC_XbpPickList:enable()

::lEnabled := .t.

::oButtonRLone:enable()
::oButtonRLall:enable()
::oButtonLRone:enable()
::oButtonLRall:enable()
IF Valtype(::oButtonOK) = 'O'
   ::oButtonOK:enable()
ENDIF

RETURN self

* ------------

METHOD DC_XbpPickList:disable()

::lEnabled := .f.

::oButtonRLone:disable()
::oButtonRLall:disable()
::oButtonLRone:disable()
::oButtonLRall:disable()
IF Valtype(::oButtonOK) = 'O'
   ::oButtonOK:disable()
ENDIF

RETURN self

* -------------

// need this to prevent DC_GetRefresh() from resetting data
// when calling :setData()

METHOD DC_XbpPickList:refresh()

::leftText:setCaption(::leftText:cargo + ' (' + ;
   Alltrim(Str(::leftListBox:numItems()) + ')'))

::rightText:setCaption(::rightText:cargo + ' (' + ;
   Alltrim(Str(::rightListBox:numItems()) + ')'))

RETURN self

* ------------

STATIC METHOD DC_XbpPickList:Keyboard( a, nMode )

IF a == xbeK_ALT_O .AND. Valtype(::oButtonOk) == 'O'
  PostAppEvent(xbeP_Activate,,,::oButtonOK)
ELSEIF SetAppFocus():isDerivedFrom('XbpPushButton') .AND. a == xbeK_ENTER
  PostAppEvent(xbeP_Activate,,,SetAppFocus())
ELSEIF a = xbeK_SPACE .OR. a = xbeK_ENTER
  IF nMode =1
    ::transferItems( ::leftListBox, ::rightListBox, .F. )
  ELSE
    ::transferItems( ::rightListBox, ::leftListBox, .F. )
  ENDIF
ENDIF

::XbpStatic:Keyboard(a)

RETURN self

* ---------------------------

METHOD DC_XbpPickList:WhenHideEval( lEnable, lShow )

IF Valtype(::whenBlock) = 'B'
  lEnable:= Eval(::whenBlock,self)
  IF lEnable
    ::enable()
  ELSE
    ::disable()
  ENDIF
ENDIF

IF Valtype(::hideBlock) = 'B'
  lShow := !Eval(::hideBlock,self)
  IF lShow
    ::show()
  ELSE
    ::hide()
  ENDIF
ENDIF

RETURN self

//////////////////////////////////////////////////////////////////////
//
//  GROUPMGR.PRG
//
//  Copyright:
//     Alaska Software Inc., (c) 1997. All rights reserved.
//
//  Contents:
//     Class GroupManager for managing the position of grouped Xbase Parts in
//     sizeable dialog windows. All XBPs of one group have the same size
//
//  Remarks:
//     A GroupManager object calculates a centered position for the
//     XBP group across a given distance, applies an offset and provides
//     for a variable space between each XBP in the group.
//     Positioning can be done in horizontal or vertical direction.
//
//     Example for horizontal arrangement:
//
//     #    ->  distance for centering
//     o    ->  offset in x direction
//     |PB| ->  Pushbuttons
//     -    ->  Variable space between pushbuttons
//
//     A) Small distance available
//
//                 ################
//           o      |PB|-|PB|-|PB|
//
//     B) Large distance available
//
//                 ####################################
//           o             |PB|----|PB|----|PB|
//
//////////////////////////////////////////////////////////////////////


#include "Appevent.ch"
#include "Common.ch"
#include "Gra.ch"
#include "Xbp.ch"


STATIC CLASS GroupManager FROM XbpPartHandler
   PROTECTED:
   VAR aXbpSize         // size of single XbasePart
   VAR aMinSize         // min size of all XbaseParts
   VAR aMaxSize         // max size of all XbaseParts
   VAR aXbps            // array with XbaseParts
   METHOD centerH       // center XbaseParts horizontally
   METHOD centerV       // center XbaseParts vertically

   EXPORTED:
   VAR maxDistance      // max. distance between XbaseParts
   VAR currentDistance  // current distance between XbaseParts
   VAR horizontal       // .T. => XbaseParts are displayed horizontally

   METHOD init
   METHOD addPart       // adds an XbasePart to :aXbps
   METHOD createButton  // creates an XbpPushbutton
   METHOD createSLE     // creates an XbpSLE
   METHOD createText    // creates an XbpStatic Type Text
   METHOD part          // returns an Xbase Part by ordinal position
   METHOD numParts      // returns the number of XbaseParts in a group
   METHOD setPos        // repositions XbaseParts group to given position
   METHOD setSize       // resizes each XbaseParts in group to same size
   METHOD center        // centers all XbaseParts within the group
   METHOD minSize       // returns minimum size of the Part group
   METHOD maxSize       // returns maximum size of the Part group
   METHOD currentPos    // returns current pos of the Part group
   METHOD currentSize   // returns current size of the Part group
ENDCLASS



/*
 * Initialize object
 */
STATIC METHOD GroupManager:init( oParent, oOwner, aSize )

   DEFAULT oParent TO SetAppWindow(), ;
           oOwner  TO oParent

   ::xbpPartHandler:init( oParent, oOwner )
   ::aXbps           := {}
   ::aXbpSize        := AClone( aSize )

   ::currentDistance := 0
   ::maxDistance     := 24
   ::horizontal      := .T.

RETURN self




/*
 * Adds a created Xbase Part to the group
 */
STATIC METHOD GroupManager:addPart( oXbp )

   AAdd( ::aXbps, oXbp )

   IF ::aMinSize == NIL
      ::aMinSize := Aclone( ::aXbpSize )
      ::aMaxSize := Aclone( ::aXbpSize )
   ELSEIF ::horizontal
      ::aMinSize[1] += ::aXbpSize[1]
      ::aMaxSize[1] += ( ::aXbpSize[1] + ::maxDistance )
   ELSE
      ::aMinSize[2] += ::aXbpSize[2]
      ::aMaxSize[2] += ( ::aXbpSize[2] + ::maxDistance )
   ENDIF

RETURN self



/*
 * Creates a new pushbutton and adds it to the group
 */
STATIC METHOD GroupManager:createButton( xCaption, bActivate, lTabstop, lPointerFocus, nGroup )
   LOCAL oButton

   DEFAULT lTabStop      TO .T. , ;
           lPointerFocus TO .T. , ;
           nGroup        TO XBP_NO_GROUP

   oButton              := XbpPushbutton():new( ::setParent(), ::setOwner(), ;
                                                NIL          , ::aXbpSize )
   oButton:activate     := bActivate
   oButton:caption      := xCaption
   oButton:group        := nGroup
   oButton:pointerFocus := lPointerFocus
   oButton:tabStop      := lTabStop
   oButton:clipSiblings := .T.
   oButton:create()

   ::addPart( oButton )

RETURN oButton



/*
 * Creates a new Sle and adds it to the group
 */
STATIC METHOD GroupManager:createSLE( xData, nBufferLength, lTabstop, nGroup )
   LOCAL oSle

   DEFAULT nBufferLength TO 32  , ;
           lTabStop      TO .T. , ;
           nGroup        TO XBP_NO_GROUP

   oSle              := XbpSLE():new( ::setParent(), ::setOwner(), ;
                                      NIL          , ::aXbpSize )
   oSle:bufferLength := nBufferLength
   oSle:tabStop      := lTabStop
   oSle:group        := nGroup
   oSle:clipSiblings := .T.
   oSle:create()

   IF Valtype( xData ) == "B"
      oSle:dataLink := xData
      oSle:setData()
   ELSE
      oSle:setData( xData )
   ENDIF

   ::addPart( oSle )

RETURN oSle



/*
 * Creates a new XbpStatic type text
 */
STATIC METHOD GroupManager:createText( cText, nHorizAlign, nVertAlign )
   LOCAL oStatic

   DEFAULT nHorizAlign TO XBPSTATIC_TEXT_LEFT , ;
           nVertAlign  TO XBPSTATIC_TEXT_VCENTER

   oStatic         := XbpStatic():new( ::setParent(), , , ::aXbpSize )
   oStatic:caption := cText
   oStatic:type    := XBPSTATIC_TYPE_TEXT
   oStatic:options := nHorizAlign + nVertAlign
   oStatic:clipSiblings := .T.
   oStatic:create()

   ::addPart( oStatic )

RETURN oStatic



/*
 * Returns an XbasePart by ordinal posintion in the group
 */
STATIC METHOD GroupManager:part( nIndex )
RETURN ::aXbps[ nIndex ]



/*
 * Returns the number of XbaseParts
 */
STATIC METHOD GroupManager:numParts()
RETURN Len( ::aXbps )



/*
 * Sets the position for the "anchor" Part and repositions all other
 * XbaseParts from there
 */
STATIC METHOD GroupManager:setPos( aPos )
   LOCAL aXbpPos, i, nCount := Len( ::aXbps )

   IF ::horizontal                // horizontal positioning left->right
      i := 1
   ELSE                           // vertical positioning bottom->top
      i := nCount
   ENDIF
   aXbpPos := AClone( aPos )
   ::aXbps[i]:setPos( aXbpPos, .F. )

   IF ::horizontal
      FOR i := 2 TO nCount
         aXbpPos[1] += ::currentDistance + ::aXbpSize[1]
         ::aXbps[i]:setPos( aXbpPos, .F. )
      NEXT
   ELSE
      nCount --
      FOR i := nCount TO 1 STEP -1
         aXbpPos[2] += ::currentDistance + ::aXbpSize[2]
         ::aXbps[i]:setPos( aXbpPos, .F. )
      NEXT
   ENDIF

   ::setParent():invalidateRect()
RETURN self



/*
 * Changes the size of all XbaseParts in the group.
 * After this, the method :center() must be called
 */
STATIC METHOD GroupManager:setSize( aSize )
   LOCAL aXbps[ ::numParts() ]

   ACopy( ::aXbps, aXbps )

   ::aXbpSize := aSize
   ::aMinSize := NIL
   ::aXbps    := {}

   AEval( aXbps, {|o| o:setSize(aSize,.F.), ::addPart(o) } )

RETURN self



/*
 * Centers the Part group across a given distance and adds an offset
 */
STATIC METHOD GroupManager:center( nWidth, nOffset )
   IF ::horizontal
      ::centerH( nWidth, nOffset )
   ELSE
      ::centerV( nWidth, nOffset )
   ENDIF
RETURN self



/*
 * Centers the group horizontally with variable distance between XbaseParts
 */
STATIC METHOD GroupManager:centerH( nWidth, nLeftOffset )
   LOCAL aPos  := ::aXbps[1]:currentPos()
   LOCAL aSize := ::currentSize()
   LOCAL nCount:= Len( ::aXbps )
   LOCAL nDist := nWidth - ( nCount * ::aXbpSize[1] )
   LOCAL i

   DEFAULT nLeftOffset TO 0

  /*
   * Variable distance between XbaseParts from 0 to :maxDistance
   */
   IF nCount > 1
      ::currentDistance := Min( ::maxDistance, Max( 0, nDist / (nCount-1) ) )
   ELSE
      ::currentDistance := 0
   ENDIF

   aPos[1] := ( nWidth - ( nCount * ::aXbpSize[1] ) - ;
                         ( ( nCount - 1 ) * ::currentDistance ) ) / 2
   aPos[1] += nLeftOffset
   aPos[1] := Max( aPos[1], nLeftOffset )

   ::aXbps[1]:setPos( aPos,.F. )
   FOR i:=2 TO nCount
      aPos[1] += ::currentDistance + ::aXbpSize[1]
      ::aXbps[i]:setPos( aPos, .F. )
   NEXT

   ::setParent():invalidateRect()
RETURN self


/*
 * Centers the group vertically with variable distance between XbaseParts
 */
STATIC METHOD GroupManager:centerV( nHeight, nBottomOffset )
   LOCAL aPos  := ::aXbps[1]:currentPos()
   LOCAL aSize := ::currentSize()
   LOCAL nCount:= Len( ::aXbps )
   LOCAL nDist := nHeight - ( nCount * ::aXbpSize[2] )
   LOCAL i

   DEFAULT nBottomOffset TO 0

  /*
   * Variable distance between XbaseParts from 0 to :maxDistance
   */
   IF nCount > 1
      ::currentDistance := Min( ::maxDistance, Max( 0, nDist / (nCount-1) ) )
   ELSE
      ::currentDistance := 0
   ENDIF

   aPos[2] := ( nHeight - ( nCount * ::aXbpSize[2] ) - ;
                          ( ( nCount - 1 ) * ::currentDistance ) ) / 2
   aPos[2] += nBottomOffset
   aPos[2] := Max( aPos[2], nBottomOffset )

   ATail(::aXbps):setPos( aPos,.F. )
   nCount --

   FOR i:= nCount TO 1 STEP -1
      aPos[2] += ::currentDistance + ::aXbpSize[2]
      ::aXbps[i]:setPos( aPos,.F. )
   NEXT

   ::setParent():invalidateRect()
RETURN self


/*
 * Position of lower left XbasePart
 */
STATIC METHOD GroupManager:currentPos
   LOCAL aPos

   IF ::horizontal
      aPos := ::aXbps[1]:currentPos()
   ELSE
      aPos := Atail(::aXbps):currentPos()
   ENDIF

RETURN aPos



/*
 * Returns the area that is occupied by all XbaseParts
 */
STATIC METHOD GroupManager:currentSize
   LOCAL aPos, aSize

   IF Len( ::aXbps ) == 1
      RETURN AClone( ::aXbpSize )      // ** RETURN **

   ELSEIF ::horizontal
      aPos  := ::aXbps[1]:currentPos()
      aSize := ATail( ::aXbps ):currentPos()
   ELSE
      aPos  := ATail( ::aXbps ):currentPos()
      aSize := ::aXbps[1]:currentPos()
   ENDIF

   aSize[1] += ::aXbpSize[1]
   aSize[2] += ::aXbpSize[2]
   aSize[1] -= aPos[1]
   aSize[2] -= aPos[2]

RETURN aSize



/*
 * The minimum area required to display all XbaseParts of the group
 */
STATIC METHOD GroupManager:minSize

IF ::aMinSize == NIL
  ::aMinSize := Aclone( ::aXbpSize )
ENDIF

RETURN AClone( ::aMinSize )



/*
 * The maximum area required to display all XbaseParts of the group
 */
STATIC METHOD GroupManager:maxSize

IF ::aMaxSize == NIL
  ::aMaxSize := Aclone( ::aXbpSize )
ENDIF

RETURN AClone( ::aMaxSize )

* -------------------
